# Application Checker
# Update module (Update.pm)
#
# Copyright (C) 2007-2010 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Update;
use strict;
use File::Temp;
use Common;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw( &db_get_version &db_check_update &db_update &date_format @db_error_msg );

sub import {
	my ($module, %params) = @_;
	
	Update->export_to_level(1, $module);
	
	init(%params);
}
#-----------------------------------------------------------------------

our $source = 'http://ftp.linuxfoundation.org/pub/lsb/updates/'.lc($appchecker_branch{'OS'}).'-appchecker-data';

our $temp_dir = undef;

our $db_ver = '6'; # KEEP IN SYNC WITH mkrawlist.pl!!!

our @db_files = qw/ Distributions IIndex Interfaces
		STD_IntAttrs STD_Ints STD_LibAttrs STD_Libs
		LibIfaces LibIndex UnapprovedLibs LDPath
		Approved Sonames RLibDeps Components/;
		# Keep in sync with update_gz.sh, package/mkrawlists.pl, package/Makefile

our $arch = 'x86';

our $dl = 'wget';
our $dl_proxy = "";
our $dl_opt = ""; # Additional options for wget/curl

#-----------------------------------------------------------------------

sub init (%) {
	my (%params) = @_;
	
	$dl_proxy = $params{PROXY} if $params{PROXY};
	$dl_opt = $params{OPTS} if $params{OPTS};
	$dl = $params{DL} if $params{DL};
	$arch = $params{ARCH} if $params{ARCH};
}
#-----------------------------------------------------------------------

our @db_error_msg = ();

# complain
our $_complain = sub {
	push @db_error_msg, @_;
};
sub complain { &$_complain(@_); return undef; }

# dbg_out
our $_dbg_out = sub {
	#if (open(LOG, '>>/tmp/1progress')) {
		#print LOG "DBG: $_[0]\n";
		#close(LOG);
	#}
};
sub dbg_out { goto &$_dbg_out; }

sub cmd {
	my ($cmd) = @_;

	dbg_out $cmd;
	return system($cmd);
}

#-----------------------------------------------------------------------

# Download file.
# &$on_file_progress($ratio) is an optional callback sub used to track download progress.
sub get_file ($$;&) {
	my ($url, $to, $on_file_progress) = @_;
	
	my $cmd = 'false';
	
	if ( $dl eq 'wget' ) {
		my $proxy_opt = '--no-proxy';
		my $proxy_pre = '';
		if ( $dl_proxy ) {
			$proxy_opt = '';
			if ( $dl_proxy =~ /^(?:([^\@:]+)(?::([^\@:]+))?\@)?([^\@]+)$/ ) {
				$1 and $proxy_opt .= " '--proxy-user=$1'";
				$2 and $proxy_opt .= " '--proxy-password=$2'";
				$proxy_pre = "http_proxy='".$3."' ftp_proxy='".$3."'";
			}
		}
		$cmd = "$proxy_pre wget --progress=bar:force -nd -np -t 1 --retr-symlinks --timeout=60 $proxy_opt $dl_opt '$url' -O '$to'";
	}
	elsif ( $dl eq 'curl' ) {
		my $proxy_opt = "--proxy ''";
		if ( $dl_proxy ) {
			$proxy_opt = '';
			if ( $dl_proxy =~ /^(?:([^\@]+)\@)?([^\@]+)$/ ) {
				$1 and $proxy_opt .= " --proxy-user '$1'";
				$proxy_opt .= " --proxy '$2'";
			}
		}
		$cmd = "curl -# --retry 1 --connect-timeout 60 $proxy_opt $dl_opt '$url' -o '$to'";
	}

	$SIG{PIPE} = 'IGNORE'; # Disable failing on SIGPIPE
	dbg_out $cmd;
    open(GET, "$cmd 2>&1 |") or return complain "Cannot run $dl: $!";
	
	my $res = 0;
	my $in = "";
	my $txt = "";
	my $prev_ratio = "";
	# Read from the pipe.
	# Can't use <GET> here, because the progress is in one line and we want to track it.
	while ( $res = read(GET, $in, 128) ) {
		$txt .= $in;
		
		while ( $txt =~ s/([^\x0A\x0D]*)[\x0A\x0D]+//s ) { # End of the line or cariage return.
			my $line = $1;
			dbg_out $line;
			if ( $dl eq 'wget' ) {
				if ( $line =~ /^\s*(\d+)%/ ) { # Ratio number is in the beginning of the line
					# Do not call again if the ration hasn't increased.
					&$on_file_progress($1) if $on_file_progress && $1 ne $prev_ratio;
					$prev_ratio = $1;
				}
			}
			elsif ( $dl eq 'curl' ) {
				if ( $line =~ /(\d+)([\.,]\d+)?%$/ ) { # Ratio number is in the end of the line
					# Do not call again if the ratio hasn't increased.
					&$on_file_progress($1) if $on_file_progress && $1 ne $prev_ratio;
					$prev_ratio = $1;
				}
			}
		}
	}
	defined $res or complain "Read error: $!"; # Pipe read error
	
	close(GET);
	$SIG{PIPE} = 'DEFAULT';
	
	defined $res or return undef; # return after the pipe has been closed
	
	if ( $? != 0 ) { # Non-zero exit code
		return complain "$dl failed".($! ? ": $!" : ". No internet connection or wrong proxy?");
	}
	
	return 1;
}

# Returns the two values:
# 1. version of the installed DB snapshots;
# 2. date/timt of the last check or 0 if there was no check before.
# It may also return <undef> if something is wrong with the current VERSIONS file.
sub db_get_version ($) {
	my ($dest_dir) = @_;
	
	my $ver_file = "VERSIONS_$arch";
	
	if ( ! -f "$dest_dir/$ver_file" ) {
		return (0, 0);
	}

	my $loc_versions = read_versions("$dest_dir/$ver_file") or return undef;
	if ( !$loc_versions->{VERSION} || $db_ver lt $loc_versions->{VERSION} ) {
		return complain "Data version is wrong (expected 'DBVERSION >= $db_ver').";
	}

	my $installed_ver = 0;
	foreach (keys %{$loc_versions->{FILES}}) {
		if ($installed_ver < $loc_versions->{FILES}->{$_}->{'VER'}) {
			$installed_ver = $loc_versions->{FILES}->{$_}->{'VER'};
		}
	}

	if ( ! $loc_versions->{LAST_CHECK} ) {
		$loc_versions->{LAST_CHECK} = 0;
	}

	return ($installed_ver, $loc_versions->{LAST_CHECK});
}

sub unlink_sure {
	my ($file) = @_;
	
	( -f $file ) or return 1; # Check if the file really exists
	dbg_out "rm '$file'";
	unlink $file   or return complain "Failed to delete '$file': $!";
	( ! -f $file ) or return complain "Failed to delete '$file'";
	
	return 1;
}

sub temp_dir () {
	my $res;
	if ( $temp_dir ) {
		$res = $temp_dir;
		( -d $temp_dir ) or return complain "Directory doesn't exists '$temp_dir'";
	}
	else { # Use a 'random' temp dir
		eval { # try
			$res = File::Temp::tempdir( 'app-checker-XXXXXX', TMPDIR => 1, CLEANUP => 1 );
		};
		if ( $@ || ! -d $res ) {
			return complain "Failed to create temp dir: $@";
		}
	}
	return $res;
}

# Check for updates
# Returns: timestamp if there is an update, "" if there is nothing to update, <undef> in case of errors.
sub db_check_update {
	my ($dest_dir) = @_;

	my $old_umask;
	if (index($dest_dir, '/var/opt/') == 0) {
		# Change default permissions on dbdata so that everyone else could overwrite new files
		$old_umask = umask(0);
	}
	
	my $ver_file = "VERSIONS_$arch";
	
	# Read the local VERSIONS file
	my $loc_versions = read_versions("$dest_dir/$ver_file") or return undef;
	if ( !$loc_versions->{VERSION} || $db_ver lt $loc_versions->{VERSION} ) {
		return complain "Data version is wrong (expected 'DBVERSION >= $db_ver').";
	}

	my $temp_dir = temp_dir() or return undef;
	
	# Delete old file
	unlink_sure("$temp_dir/$ver_file.new") or return undef;
	
	# Download a new VERSIONS file
	get_file("$source/$ver_file", "$temp_dir/$ver_file.new")
		or return complain "Failed to download '$ver_file'";
	
	(-f "$temp_dir/$ver_file.new") or return complain "Failed to download '$ver_file'";
	
	# Read the new VERSIONS file
	my $new_versions = read_versions("$temp_dir/$ver_file.new")	or return undef;
	if ( !$new_versions->{VERSION} || $db_ver lt $new_versions->{VERSION} ) {
		return complain "Installed Application Checker is too old to use the latest updates. Database structure has changed. (New DB ver is '".$new_versions->{VERSION}."').";
	}
	
	# Select max version among new files.
	my $max_version = "";
	# Calc download size
	my $dl_size = 0;
	foreach my $db_file ( @db_files ) { # For all possible file names
		my $file = "${db_file}_${arch}.gz";
		my $loc_file = $loc_versions->{FILES}->{$file};
		my $new_file = $new_versions->{FILES}->{$file};
		if ( $new_file
			&& ( !$loc_file || $new_file->{VER} gt $loc_file->{VER}) )
		{
			$max_version = $new_file->{VER} if $new_file->{VER} gt $max_version;
			$dl_size += $new_file->{SIZE};
		}
	}
	
	# Update the LAST_CHECK field in the current VERSIONS file.
	$loc_versions->{LAST_CHECK} = timestamp( localtime(time) );
	
	write_versions($loc_versions, "$dest_dir/$ver_file")
		or complain "Failed to save time of the last check.";
	
	if ($old_umask) {
		# Restore umask if it was changes
		umask($old_umask);
	}
	
	# $max_version is "" if there is nothing to update
	return ($max_version, $dl_size);
}

# Download updated database snapshots
# &$progress_callback($$$) is a callback sub used to track the download progress. See description of parameters below where it's used.
# Returns: 1 if updated OK, 0 if there is nothing to update, <undef> in case of errors.
sub db_update ($&) {
	my ($dest_dir, $progress_callback) = @_;
	
	my $old_umask;
	if (index($dest_dir, '/var/opt/') == 0) {
		# Change default permissions on dbdata so that everyone else could overwrite new files
		$old_umask = umask(0);
	}
	
	my $ver_file = "VERSIONS_$arch";
	
	# Read the current VERSIONS file
	my $loc_versions = read_versions("$dest_dir/$ver_file")	or return undef;
	if ( !$loc_versions->{VERSION} || $db_ver lt $loc_versions->{VERSION} ) {
		return complain "Data version is wrong (expected 'DBVERSION >= $db_ver').";
	}
	
	my $temp_dir = temp_dir() or return undef;
	
	# Delete old file
	unlink_sure("$temp_dir/$ver_file.new") or return undef;
	
	# Download
	get_file("$source/$ver_file", "$temp_dir/$ver_file.new")
		or return complain "Failed to download '$ver_file'";
	
	(-f "$temp_dir/$ver_file.new") or return complain "Failed to download '$ver_file'";
	
	# Read new VERSIONS file
	my $new_versions = read_versions("$temp_dir/$ver_file.new")	or return undef;
	if ( !$new_versions->{VERSION} || $db_ver lt $new_versions->{VERSION} ) {
		return complain "Installed Application Checker is too old to use the latest updates. Database structure has changed. (New DB ver is '".$new_versions->{VERSION}."').";
	}
	
	my @to_dl = (); # "to download" list
	
	foreach my $db_file ( @db_files ) { # For all possible file names
		my $file = "${db_file}_${arch}.gz";
		my $loc_file = $loc_versions->{FILES}->{$file};
		my $new_file = $new_versions->{FILES}->{$file};
		
		if ( $new_file
			&& ( !$loc_file || $new_file->{VER} gt $loc_file->{VER}) )
		{
			# A newer version. To be downloaded.
			push @to_dl, $file;
			
			# Replace an old description with a new one
			$loc_versions->{FILES}->{$file} = $new_versions->{FILES}->{$file};
		}
	}
	
	my @to_replace = ();
	
	# Calc total download size
	my $dl_size = 0;
	foreach my $file ( @to_dl ) {
		$dl_size += $loc_versions->{FILES}->{$file}->{SIZE};
	}
	my $downloaded = 0;
	
	foreach my $file ( @to_dl ) {
		# Aux multipliers
		my $a = $downloaded / $dl_size * 100;
		my $b = $loc_versions->{FILES}->{$file}->{SIZE} / $dl_size;
		
		# This callback is called during the one file download progress
		my $on_file_progress = sub {
				my ($ratio) = @_; # Current file download ratio
				
				return if ( $ratio < 0 || $ratio >100 ); # Keep silence if there are problems.
				
				# Calc the total download progress
				my $total = int( $a + $b * $ratio );
				
				# Call the given callback function. Parameters:
				#  total ratio (0-100%),
				#  the file being downloaded (filename),
				#  Current file download ratio (0-100%).
				&$progress_callback( $total, $file, $ratio);
			};
		
		&$on_file_progress(0); # 'Get ready' notification
		
		# Download the file:
		unlink_sure("$temp_dir/$file") or return undef;
		get_file( "$source/$file", "$temp_dir/$file", \&$on_file_progress )
			or return complain "Failed to load '$file'";
		
		&$on_file_progress(100); # In case that wget didn't print progress for any reason.
		
		# Check the downloaded file
		-f "$temp_dir/$file" or return complain "Failed to load '$file'";
		(-s "$temp_dir/$file") == $loc_versions->{FILES}->{$file}->{SIZE}
			or return complain "Size mismatch: '$file'";
		
		push @to_replace, "$temp_dir/$file";
		
		$downloaded += $loc_versions->{FILES}->{$file}->{SIZE};
	}
	
	# OK. Now all the downloaded and unpacked files should be recorded in the @to_replace array.
	
	$loc_versions->{VERSION} = $new_versions->{VERSION};
	$loc_versions->{LAST_CHECK} = timestamp( localtime(time) );
	
	if ( @to_replace ) {
		# The VERSIONS file should be also replaced.
		# Make a new one
		unlink_sure("$temp_dir/$ver_file") or return undef;
		write_versions($loc_versions, "$temp_dir/$ver_file")
			or return complain "Failed to create a new VERIONS file: '$temp_dir/$ver_file'";
		
		# And add it to the "to be replaced" list.
		push @to_replace, "$temp_dir/$ver_file";
		
		# Do the replacment in a single step
		cmd("mv -f ".(join " ", map {"'$_'"} @to_replace)." '$dest_dir'") == 0
			or return complain "Failed to replace old files";
	} else {
		# Nothing to replace. Files are up to date.
		
		# Update the LAST_CHECK timestamp
		write_versions($loc_versions, "$dest_dir/$ver_file")
			or complain "Failed to save time of the last check.";
		
		return 0; # Everything is already up to date.
	}
	
	if ($old_umask) {
		# Restore umask if it was changes
		umask($old_umask);
	}
	
	# Updated successfully
	return 1;
}

# Write the VERSIONS file
# Returns: 1 if OK, <undef> if errors
sub write_versions ($$) {
	my ($versions, $filename) = @_;
	
	open FILE, ">$filename"
		 or return complain "Can't open file for writing '$filename': $!";
	
	print FILE "DBVERSION >= ".($versions->{VERSION} or $db_ver)."\n";
	print FILE "LAST_CHECK ".$versions->{LAST_CHECK}."\n" if $versions->{LAST_CHECK};
	print FILE "# Filename\tTimestamp\tSize\n";
	
	foreach my $file ( sort keys %{$versions->{FILES}} ) {
		my $data = $versions->{FILES}->{$file};
		print FILE join "\t", $file, $data->{VER}, $data->{SIZE};
		print FILE "\n";
	}
	
	close FILE;
	
	return 1;
}

# Read the VERSIONS file
# Returns: HASHREF struct if read OK, <undef> if there were errors
sub read_versions ($) {
	my ($filename) = @_;
	
	my $result = { FILES => {} };
	
	dbg_out "Reading ver_file '$filename'";
	open FILE, $filename or return $result; # The file may be absent
	
	while ( my $line = <FILE> ) {
		chomp $line;
		next if $line =~ /^\s*$/; # skip empty lines
		next if $line =~ /^#/; # skip comments

		if ( $line =~ /^DBVERSION\s*>=\s*(.+)/ ) {
			$result->{VERSION} = $1;
		}
		elsif ( $line =~ /^LAST_CHECK\s+(.*)/ ) {
			$result->{LAST_CHECK} = $1;
		}
		# Format:  filename  timestamp  size\n
		elsif ( $line =~ /^([^\t]+)\t([^\t]+)\t([^\t]+)$/ ) {
			my ($file, $t_ver, $size) = ($1, $2, $3);
			$result->{FILES}->{$file} = { VER => $t_ver, SIZE => $size};
		}
		else {
			return complain "Wrong syntax at $filename:$.";
		}
	}
	
	close FILE;
	
	return $result;
}

sub timestamp (@) {
	my ($sec, $min, $hour, $mday, $mon, $year) = @_;
	$year += 1900; $mon += 1;
	return sprintf("%i%02i%02i%02i%02i%02i", $year, $mon, $mday, $hour, $min, $sec);
}

# YYYYMMDDhhmmss -> DD.MM.YYYY hh:mm:ss
sub date_format($) {
	my ($timestamp) = @_;
	my $res = $timestamp;
	if ($timestamp =~ m/^(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})$/) {
		$res = "$3.$2.$1 $4:$5:$6";
	}
	return $res;
}
