#!/usr/bin/perl -w

# Application Checker
# Configuration Save Module (tests_conf_save.pl)
#
# Copyright (C) 2007-2010 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use Cwd qw(cwd);
use Fcntl ':flock';
use Fcntl ':seek';
use File::Temp qw/tempfile tmpnam/;
use FindBin;

sub BEGIN { push @INC, $FindBin::Bin; }

use Common;

autoflush_on();

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STARTUP SECTION

my $default_spec_version = (($appchecker_branch{'STD'} eq 'LSB') ? '4.1' : '2.0');

# Prints information about command-line parameters
sub usage() {
	my $spec_params = '';
	if ($appchecker_branch{'STD'} eq 'LSB') {
		$spec_params = <<DATA;
    --spec=<LSB-version>     LSB version to test against ($default_spec_version by default).
    -T [core,c++|core,c++,desktop]
                             LSB profile to test against ("Core & C++ & Desktop"
                             by default).
DATA
	}
	else {
		$spec_params = <<DATA;
    --spec=<Moblin-version>  Moblin version to test against ($default_spec_version by default).
DATA
	}
	print <<DATA;
    Usage: $0 [OPTION]...
    At least one of the --profile, --paths and --packages must be specified!

    --paths=<paths-list>     List of files and directories to be tested. For
                             directories all their contents is taken. The list
                             is delimited by a colon ":".
    --packages=<packages-list>
                             List of installed packages (rpm or deb) which
                             contents is to be tested. The list is delimited by
                             a colon ":".
    -n <name>, --name=<name> Name of the application. If not specified, the name
                             of the first file/directory/package will be taken.
$spec_params    --comment=<text>         Comments to the test run. Several comments being
                             provided, they will be joined.
    -f <file>, --profile=<file>
                             Alternatively to specifying all the options above,
                             you may provide a text file containing all of them
                             inside. The profile can be created beforehand by
                             means of the AppChecker Web-UI.
    -r <dir>, --results-dir=<dir>
                             Directory for storing the results.
    -t <dir>, --tests-dir=<dir>
                             Directory containing individual check tools.
    -d <dir>, --dbdata-dir=<dir>
                             Directory containing database snapshots.
    -m [rpm|dpkg]            Force using the specified package manager (RPM or
                             dpkg).
    --verbose                Output information about test stages.
    -h, --help               Display this help and exit.
    -v, --version            Output version information and exit.
DATA
}

my $error;

# Parameters for testing (received via command line or with default values)
my @paths = ();                                 # Files/directories to be tested (--paths)
my @packages = ();                              # Packages to be tested (--packages)
my $app_name = '';                              # Application name (--name)
my $spec_version = $default_spec_version;       # Spec version (--spec)
my $spec_version_specified = 0;
my $spec_profile = 'core,c++,desktop';          # Spec profile (-T), for LSB only
my $spec_profile_specified = 0;
my $comments = '';                              # Comments to the test run (--comment)
my $profile = '';                               # Profile (--profile)
my $results_path = '';                          # 'results' directory (--results-dir)
my $results_path_specified = 0;
my $tests_path = '';                            # 'bin' directory (--tests-dir)
my $tests_path_specified = 0;
my $dbdata_path = '';                           # 'dbdata' directory (--dbdata-dir)
my $dbdata_path_specified = 0;
my $package_manager = '';                       # Package manager to be used (-m)
my $verbose = 0;                                # Output test stages (--verbose)
my $show_help = 0;                              # Display help (--help)
my $show_version = 0;                           # Display version (--version)

my $utils_path = $FindBin::Bin;                 # 'utils' path

$tests_path = abs_path("$utils_path/../bin");
if ($utils_path eq "/opt/$appchecker_branch{'DIR'}/app-checker/utils") {
	$dbdata_path = "/var/opt/$appchecker_branch{'DIR'}/app-checker/dbdata";
	if (getpwuid($>) eq $appchecker_branch{'USER'}) {       # Running with special user permissions, use global directory
		$results_path = "/var/opt/$appchecker_branch{'DIR'}/app-checker/results";
	}
	else {
		$results_path = (getpwuid($>))[7]."/.$appchecker_branch{'DIR'}/app-checker/results";
	}
	if ($appchecker_branch{'STD'} eq 'LSB') {
		$tests_path = "/opt/$appchecker_branch{'DIR'}/bin";
	}
}
else {
	$dbdata_path = abs_path("$utils_path/../dbdata");
	$results_path = abs_path("$utils_path/../results");
}

my $appchk_file = $tests_path.'/'.lc($appchecker_branch{'STD'}).'appchk';
my $architecture = guess_test_arch($appchk_file);           # Architecture name (unified)
my $f_architecture = lc($architecture);                     # Architecture name used in file names

# Reads next command line parameter.
# Parameters:
#   1. (IN)  Ref to array of hashes containing info about options: {LONG_NAME => 'param', SHORT_NAME => 'p', HAS_VALUE => [1|0]}
#   2. (OUT) Ref to array with long name of the parameter (or short, if not present) and its value (undef if not needed).
#      If parameter is invalid, this parameter receives (undef, undef).
# Returns 1 if the parameter has been read successfully, 0 otherwise.
sub getoption($$) {
	my ($options, $result) = @_;
	my $key;
	my $val;
	my $param = shift @ARGV;
	if (!$param) {
		return 0;
	}
	foreach (@$options) {
		my $long_name = $_->{'LONG_NAME'};
		my $short_name = $_->{'SHORT_NAME'};
		my $return_name = ($long_name or $short_name);
		if ($_->{'HAS_VALUE'}) {
			if ($long_name and ($param =~ m/^--\Q$long_name\E=(.+)$/)) {
				@$result = ($return_name, $1);
				return 1;
			}
			elsif ($short_name) {
				if ($param =~ m/^-\Q$short_name\E(.+)$/) {
					@$result = ($return_name, $1);
					return 1;
				}
				elsif ($param =~ m/^-\Q$short_name\E$/) {
					my $value = shift @ARGV;
					if (!$value) {
						print "    Error: Option $param must be followed by a value!\n";
						$error = 1;
						@$result = (undef, undef);
					}
					else {
						@$result = ($return_name, $value);
					}
					return 1;
				}
			}
		}
		else {
			if ($long_name and ($param =~ m/^--\Q$long_name\E$/)) {
				@$result = ($return_name, undef);
				return 1;
			}
			elsif ($short_name and ($param =~ m/^-\Q$short_name\E$/)) {
				@$result = ($return_name, undef);
				return 1;
			}
		}
	}
	print "    Error: Unknown option: $param\n";
	$error = 1;
	@$result = (undef, undef);
	return 1;
}

my @cmd_line_options = (
	{'LONG_NAME' => 'paths',       'SHORT_NAME' => undef, 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'packages',    'SHORT_NAME' => undef, 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'name',        'SHORT_NAME' => 'n',   'HAS_VALUE' => 1},
	{'LONG_NAME' => 'spec',        'SHORT_NAME' => undef, 'HAS_VALUE' => 1},
	{'LONG_NAME' => undef,         'SHORT_NAME' => 'T',   'HAS_VALUE' => 1},
	{'LONG_NAME' => 'comment',     'SHORT_NAME' => undef, 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'profile',     'SHORT_NAME' => 'f',   'HAS_VALUE' => 1},
	{'LONG_NAME' => 'results-dir', 'SHORT_NAME' => 'r',   'HAS_VALUE' => 1},
	{'LONG_NAME' => 'tests-dir',   'SHORT_NAME' => 't',   'HAS_VALUE' => 1},
	{'LONG_NAME' => 'dbdata-dir',  'SHORT_NAME' => 'd',   'HAS_VALUE' => 1},
	{'LONG_NAME' => undef,         'SHORT_NAME' => 'm',   'HAS_VALUE' => 1},
	{'LONG_NAME' => 'verbose',     'SHORT_NAME' => undef, 'HAS_VALUE' => 0},
	{'LONG_NAME' => 'help',        'SHORT_NAME' => 'h',   'HAS_VALUE' => 0},
	{'LONG_NAME' => 'version',     'SHORT_NAME' => 'v',   'HAS_VALUE' => 0}
);

# Process command line
my @param = ();
while (getoption(\@cmd_line_options, \@param)) {
	next if (!$param[0]);
	if ($param[0] eq 'paths') {
		@paths = map(abs_path($_), split(/:/, $param[1]));
	}
	elsif ($param[0] eq 'packages') {
		@packages = split(/:/, $param[1]);
	}
	elsif ($param[0] eq 'name') {
		$app_name = $param[1];
	}
	elsif ($param[0] eq 'spec') {
		$spec_version = $param[1];
		$spec_version_specified = 1;
	}
	elsif (($param[0] eq 'T') and ($appchecker_branch{'STD'} eq 'LSB')) {
		$spec_profile = $param[1];
		$spec_profile_specified = 1;
	}
	elsif ($param[0] eq 'comment') {
		$comments .= (($comments eq '') ? '' : "\n").$param[1];
	}
	elsif ($param[0] eq 'profile') {
		$profile = $param[1];
	}
	elsif ($param[0] eq 'results-dir') {
		$results_path = abs_path($param[1]);
		$results_path_specified = 1;
	}
	elsif ($param[0] eq 'tests-dir') {
		$tests_path = abs_path($param[1]);
		$tests_path_specified = 1;
	}
	elsif ($param[0] eq 'dbdata-dir') {
		$dbdata_path = abs_path($param[1]);
		$dbdata_path_specified = 1;
	}
	elsif ($param[0] eq 'm') {
		$package_manager = $param[1];
	}
	elsif ($param[0] eq 'verbose') {
		$verbose = 1;
	}
	elsif ($param[0] eq 'help') {
		$show_help = 1;
	}
	elsif ($param[0] eq 'version') {
		$show_version = 1;
	}
}

if ($error) {                           # There were problems with command-line parameters
	print "    Run $0 --help for more information.\n";
	exit 1;
}

if ($show_help) {
	usage();
	exit 0;
}

if ($show_version) {
	print <<DATA;
$appchecker_branch{'OS'} Application Checker version $appchecker_ver
Copyright (C) 2007-2010 The Linux Foundation. All rights reserved.

DATA
	exit 0;
}

# If profile is specified, read it
if ($profile) {
	if (!open(FILE, $profile)) {
		print "    ERROR: Could not open file $profile\nERROR: $!\n";
		exit 2;
	}
	while (<FILE>) {
		s/[\r\n]//g;
		if (m/^APPLICATION:\s*(.*)/i) {
			$app_name = $1;
		}
		elsif (m/^COMPONENT:\s*(.*)/i) {
			my $comp = $1;
			if ($comp =~ /^pkg:(.*)/) {
				push @packages, $1;
			}
			else {
				push @paths, abs_path($comp);
			}
		}
		elsif (m/^ARCHITECTURE:\s*(.*)/i) {
			$architecture = $1;
			$f_architecture = lc($architecture);
		}
		elsif (m/^SPEC_VERSION:\s*(.*)/i) {
			$spec_version = $1;
		}
		elsif ((m/^SPEC_PROFILE:\s*(.*)/i) and ($appchecker_branch{'STD'} eq 'LSB')) {
			$spec_profile = $1;
		}
		elsif (m/^COMMENTS: (.*)/i) {
			$comments .= (($comments eq '') ? '' : "\n").$1;
		}
	}
	close(FILE);
}

# Check all parameters for consistency and fill in the default values when necessary
if (!$package_manager) {
	$package_manager = ($is_packager_rpm ? 'rpm' : 'dpkg');
}
else {
	if ((lc($package_manager) ne 'rpm') and (lc($package_manager) ne 'dpkg')) {
		print "    Error: Unknown package manager '$package_manager' specified!\n";
		print "    Run $0 --help for more information.\n";
		exit 1;
	}
}

if (($appchecker_branch{'STD'} eq 'LSB') and (lc($spec_profile) ne 'core,c++') and (lc($spec_profile) ne 'core,c++,desktop')) {
	print "    Error: Unknown $appchecker_branch{'STD'} profile '$spec_profile' specified!\n";
	print "    Run $0 --help for more information.\n";
	exit 1;
}

if ((scalar(@paths) == 0) and (scalar(@packages) == 0)) {
	print "    Error: There is nothing to test!\n";
	print "    Run $0 --help for more information.\n";
	exit 1;
}

my $spec_profiles_list = '';
if ($appchecker_branch{'STD'} eq 'LSB') {
	$spec_profiles_list =  '('.join(' & ', map(ucfirst, split(/,/, $spec_profile))).')';
}
print "    Using $appchecker_branch{'STD'} $spec_version$spec_profiles_list\n";
print "    Results directory is not specified, using $results_path\n" if (!$results_path_specified);
print "    Tests directory is not specified, using $tests_path\n" if (!$tests_path_specified);
print "    Database snapshots directory is not specified, using $dbdata_path\n" if (!$dbdata_path_specified);
print "\n";

if (!$app_name) {
	if (scalar(@paths) > 0) {
		$app_name = $paths[0];
		$app_name =~ m!/([^/]+)/?$!;
		$app_name = $1;
	}
	else {
		$app_name = $packages[0];
	}
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# PREPARE DATA SECTION

my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime();
my $host_name = `uname -n`;
$host_name =~ s/[\r\n]//g;

# Directory name for this specific test run
my $results_dir = sprintf("$f_architecture-$host_name-$app_name-%04d-%02d-%02d-%02dh-%02dm-%02ds", ($year + 1900), ($mon + 1), $mday, $hour, $min, $sec);

# Output file
my $tests_output = "$results_path/$results_dir/test_logs/tests_output";

# File with collected application analysis data
my $upload_data_file = "$results_path/$results_dir/${app_name}_${f_architecture}_upload_data";

# Detects type of the file
sub detect_file_type($) {
	my ($path) = @_;
	if (!-e $path) {
		return 'error:File not found';
	}
	if (-d $path) {
		return 'dir';       # Directory
	}
	if (!-f $path) {
		return 'error:not a file';
	}

	if (($path =~ m/\.tgz$/) or ($path =~ m/\.tar\.gz$/)) {
		return 'tgz';
	}
	elsif (($path =~ m/\.tbz$/) or ($path =~ m/\.tbz2$/) or ($path =~ m/\.tb2$/) or ($path =~ m/\.tar\.bz2$/)) {
		return 'tbz';
	}
	elsif ($path =~ m/\.tar$/) {
		return 'tar';
	}
	elsif ($path =~ m/\.rpm$/) {
		return 'rpm';
	}
	elsif ($path =~ m/\.deb$/) {
		return 'deb';
	}
	elsif ($path =~ m/\.desktop$/) {
		return 'desktop';
	}
	elsif (($path =~ m/\.png$/) or ($path =~ m/\.svg$/)) {
		return 'icon';
	}

	# Check the first line for interpreter specification ("#!...")
	my $f;
	my $first_line = '';
	open $f, $path or return "error:$!";
	read $f, $first_line, 127; # 127 bytes should be enough
	close $f;

	if ($first_line && $first_line =~ m/^#!/) {
		my $interpreter = '';
		if ($first_line =~ m|^#!\s*\S*/([^/\s]+)|) {
			$interpreter = lc($1);
			return 'perl' if ($interpreter eq 'perl');
			return 'python' if ($interpreter eq 'python');
			return 'shell' if ($interpreter =~ m/^(a|ba|da|z|k|c|tc|)sh$/);
		}
		return "script:$interpreter";   # Unknown script interpreter
	}
	if ($first_line && $first_line =~ m/^\x7fELF/) { # 0x7f ELF - Magic identificator
		my $enc = ord(substr($first_line, 5, 1)); # encoding: 1 - LSB(little-endian); 2 - MSB(big-endian)
		unless ($enc == 1 || $enc == 2) {
			return 'elf-unsupported';
		}

		my $type = substr($first_line, 16, 2); # read 2 bytes both for for 32bit and for 64bit
		if ($enc == 2) {       # reverse order of bytes
			$type = scalar reverse $type;
		}
		$type = ord(substr($type, 0, 1)); # type is a little numer, so it's ok to read only one byte of two.

		if ($type == 2) {             # Executable
			return 'elf-bin';
		}
		elsif ($type == 3) {          # Shared object
			return 'elf-so';
		}
		else {
			return 'elf-unsupported';                                           # Unsupported ELF type
		}
	}

	if ($path =~ m/\.(pl|pm)$/) { return 'perl'; };      # Perl module
	if ($path =~ m/\.py$/) { return 'python'; };         # Python module

	return 'unknown';
}

# Constructs for the specific file a unique journal name compatible with appchk's ones.
# Returns array: first element is the journal name generated, second - HTML anchor name for it.
sub get_journal_suffix_anchor($) {
	my ($fpath) = @_;
	if ($fpath =~ m/^PKG:\s*(.*)/) {
		my $pkg_name = $1;
		return ("$pkg_name.0-0.journal", "${pkg_name}_0_0");
	}
	else {
		my $fname = $fpath;
		$fname =~ s|^.*/([^/]*)$|$1|;
		my @tmp = stat($fpath);
		my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = ((scalar(@tmp) > 0) ? @tmp : (0)x13);
		my $fname_anchor = esc_id($fname);
		return ("$fname.$dev-$ino.journal", "${fname_anchor}_${dev}_${ino}");
	}
}

# Creates small journal just for indicating single PASS or FAIL.
sub create_journal($$$;$) {
	my ($path, $verdict, $type, $msg) = @_;
	my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime();
	$year += 1900;
	$mon++;
	my $tm = sprintf('%02d:%02d:%02d', $hour, $min, $sec);
	my $date = sprintf('%04d%02d%02d', $year, $mon, $mday);
	if ($msg) {
		$msg = join("\n", map {$_ = "520|0 1 $tm|$_"} split(/\n/, $msg));
	}
	my $result = ($verdict ? "$msg\n220|0 1 $verdict $tm|FAIL" : "220|0 1 0 $tm|PASS");
	my $login = getpwuid($<);
	my $uname = `uname -snrvm`;
	$uname =~ s/[\r\n]//g;
	my ($suffix) = get_journal_suffix_anchor($path);
	my $file_name = "$results_path/$results_dir/test_journals/".lc($appchecker_branch{'STD'}).$type."chk.$suffix";
	if (open(JOURNAL_FILE, ">$file_name")) {
		print JOURNAL_FILE <<DATA;
0|3.7-lite $tm $date|User: $login, Command line: <none>
5|$uname|System Information
10|0 $path $tm|TC Start
15|0 tetj-1.0 1|TCM Start
400|0 1 1 $tm|IC Start
200|0 1 $tm|TP Start, Check file type
$result
410|0 1 1 $tm|IC End
80|0 0 $tm|TC End
900|$tm|TCC End
DATA
		close(JOURNAL_FILE);
	}
	else {
		print "    Error: Failed to create file $file_name\n    $!";
	}
}

sub create_journal_INFO($$$) {
	my ($path, $title_path, $type) = @_;
	my ($suffix, $anchor) = get_journal_suffix_anchor($path);
	my $journal_file_prefix = lc($appchecker_branch{'STD'}).$type.'chk';
	my $file_name = "$results_path/$results_dir/test_journals/$journal_file_prefix.$suffix.INFO";
	if (!-f $file_name) {
		if (open(INFO_FILE, ">$file_name")) {
			print INFO_FILE "TITLE: $title_path\nANCHOR: ${journal_file_prefix}_$anchor\nJOURNAL_FILENAME: $journal_file_prefix.$suffix\n";
			close(INFO_FILE);
			print "PROGRESS: Add: JOURNAL.\n" if ($verbose);
		}
		else {
			print "    Error: Failed to create file $file_name\n    $!";
		}
	}
	return "$journal_file_prefix.$suffix";
}

# Looks for the unique temp directory name, creates it and returns its path
my $idx = 0;
sub create_temp_dir(;$) {
	my $prefix = ($_[0] or 'appcheck-temp');
	while (-e "/tmp/$prefix.$$-$idx") {
		++$idx;
	}
	mkdir("/tmp/$prefix.$$-$idx");
	return "/tmp/$prefix.$$-$idx";
}

# file_path => journal_name;
my %journals = ();

# pkg_path => unpacked_dir (for Moblin AppChecker's pkgchk)
my %unpacked_pkgs = ();

# File lists for testing with specific test programs
my @list_appchk = ();               # appchk
my @list_appchk_sh = ();            # appchk-shell (Linux AppChecker only)
my @list_appchk_pl = ();            # appchk-perl
my @list_appchk_py = ();            # appchk-python
my @list_pkgchk = ();               # pkgchk
my @list_desktopchk = ();           # desktopchk (Moblin AppChecker only)
my @list_desktopchk_icon = ();      # desktopchk icons (Moblin AppChecker only)

my @list_db = ();           # File list to be passed to dependency analyser
my @clean_dirs = ();        # List of temp directories to be cleaned after execution

my @failed_files = ();      # List of files that caused problems

# Adds directory contents to the list (recursively)
sub add_dir_to_list($$);    # Prototype for recursive function
sub add_dir_to_list($$) {
	my ($dir, $file_list) = @_;
	if (opendir(DIR, $dir)) {
		my @elems = sort(grep(!/^\.+$/, readdir(DIR)));
		closedir(DIR);
		foreach (@elems) {
			if (-f "$dir/$_") {
				$file_list->{"$dir/$_"} = 1;
			}
			else {
				add_dir_to_list("$dir/$_", $file_list);
			}
		}
	}
}

# Reads the directory contents and process every file from it (recursively)
sub process_dir($$) {
	my ($path, $title_path) = @_;
	my %file_list = ();
	add_dir_to_list($path, \%file_list);
	foreach my $file (sort keys %file_list) {
		my $title = $file;
		$title =~ s/^\Q$path\E/$title_path|/;
		process_file($file, $title, 0);
	}
}

# Creates INFO file for the future journal and determines the test program for the file
# Input parameters:
# $path       - Real path to the file under test.
# $title_path - Visible path (including package from which the file or its parent package was unpacked).
# $root       - 1 if the file/package was specified explicitly by the user; 0 otherwise.
sub process_file($$$) {
	my ($path, $title_path, $root) = @_;
	my $temp_dir;
	my $type = detect_file_type($path);
	if ($type eq 'dir') {
		process_dir($path, $title_path);
	}
	elsif ($type eq 'tgz') {
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system(esc_cmdline("echo \"Unpacking TGZ package $path\" >>«$tests_output»"));
		system(esc_cmdline("cd «$temp_dir» && tar xzf «$path» --force-local 2>>«$tests_output»"));
		if (!$?) {
			process_dir($temp_dir, $title_path);
		}
		else {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 1, 'app', 'Failed to unpack the archive.');
			}
		}
	}
	elsif ($type eq 'tbz') {
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system(esc_cmdline("echo \"Unpacking TBZ package $path\" >>«$tests_output»"));
		system(esc_cmdline("cd «$temp_dir» && tar xjf «$path» 2>>«$tests_output»"));
		if (!$?) {
			process_dir($temp_dir, $title_path);
			$journals{$path} = create_journal_INFO($path, "$title_path|", 'app');
			create_journal($path, 101, 'app', 'TAR.BZ2 archives are not standardized by the '.$appchecker_branch{'STD'}.'.');
		}
		else {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 1, 'app', 'Failed to unpack the archive.');
			}
		}
	}
	elsif ($type eq 'tar') {
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system(esc_cmdline("echo \"Unpacking TAR package $path\" >>«$tests_output»"));
		system(esc_cmdline("cd «$temp_dir» && tar xf «$path» 2>>«$tests_output»"));
		if (!$?) {
			process_dir($temp_dir, $title_path);
		}
		else {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 1, 'app', 'Failed to unpack the archive.');
			}
		}
	}
	elsif ($type eq 'rpm') {
		$journals{$path} = create_journal_INFO($path, "$title_path|", 'pkg');
		push @list_pkgchk, $path;
		print "PROGRESS: Add: PKG.\n" if ($verbose);
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system(esc_cmdline("echo \"Unpacking RPM package $path\" >>«$tests_output»"));
		my $fail = 0;
		# First extract the CPIO archive
		my $cpio_file = tmpnam();
		system(esc_cmdline("touch «$cpio_file»"));
		system(esc_cmdline("cd «$temp_dir» && rpm2cpio «$path» >«$cpio_file» 2>>«$tests_output»"));
		if (!$?) {
			# Try to unpack it as CPIO
			system(esc_cmdline("cd «$temp_dir» && cat «$cpio_file» 2>>«$tests_output» | cpio -id 2>>«$tests_output»"));
			if ($?) {
				# Failed, probably it's an undetected LZMA-packed CPIO
				if (system('which lzma') == 0) {
					system(esc_cmdline("echo «Trying to unpack using LZMA...» >>«$tests_output»"));
					system(esc_cmdline("cd «$temp_dir» && cat «$cpio_file» 2>>«$tests_output» | lzma -d 2>>«$tests_output» | cpio -id 2>>«$tests_output»"));
					if ($?) {
						$fail = 1;
					}
				}
				else {
					# No lzma is present, cannot do anything. Report the error and exit.
					$fail = 1;
				}
			}
		}
		else {
			$fail = 1;
		}
		unlink($cpio_file);
		if (!$fail) {
			if ($appchecker_branch{'STD'} eq 'LSB') {
				# For LSB get pre/post-install scripts for checking them with appchk-shell
				system(esc_cmdline("cd «$utils_path» && ./pkg-scripts.pl -u RPM -d «$temp_dir/SCRIPTS» «$path» 2>>«$tests_output»"));
			}
			else {
				# For Moblin there is no appchk-shell, but package-check needs the package unpacked
				$unpacked_pkgs{$path} = $temp_dir;
			}
			process_dir($temp_dir, $title_path);
		}
		else {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'pkg');
				create_journal($path, 1, 'pkg', 'Failed to unpack the archive.');
			}
		}
	}
	elsif ($type eq 'deb') {
		if ($appchecker_branch{'STD'} ne 'LSB') {
			# Moblin AppChecker's pkgchk can check DEB packages
			$journals{$path} = create_journal_INFO($path, "$title_path|", 'pkg');
			push @list_pkgchk, $path;
			print "PROGRESS: Add: PKG.\n" if ($verbose);
		}
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system(esc_cmdline("echo \"Unpacking DEB package $path\" >>«$tests_output»"));
		system(esc_cmdline("cd «$temp_dir» && ar -x «$path» 2>>«$tests_output»"));
		if ($?) {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'pkg');
				create_journal($path, 1, 'pkg', 'Failed to unpack the archive.');
			}
		}
		else {
			if ((-f "$temp_dir/data.tar.gz") or (-f "$temp_dir/data.tar.bz2")) {
				$temp_dir2 = create_temp_dir('pkg-unpack');
				push @clean_dirs, $temp_dir2;
				if (-f "$temp_dir/data.tar.gz") {
					system(esc_cmdline("cd «$temp_dir2» && tar xzf «$temp_dir/data.tar.gz» --force-local 2>>«$tests_output»"));
				}
				else {
					system(esc_cmdline("cd «$temp_dir2» && tar xjf «$temp_dir/data.tar.bz2» --force-local 2>>«$tests_output»"));
				}
				if (!$?) {
					if ($appchecker_branch{'STD'} eq 'LSB') {
						# For LSB get pre/post-install scripts for checking them with appchk-shell
						system(esc_cmdline("cd «$utils_path» && ./pkg-scripts.pl -u deb -d «$temp_dir2/SCRIPTS» «$path» 2>>«$tests_output»"));
						# Create journal with single warning that DEBs are not standardized
						$journals{$path} = create_journal_INFO($path, "$title_path|", 'pkg');
						create_journal($path, 101, 'pkg', 'DEB archives are not standardized by the LSB.');
					}
					else {
						# For Moblin there is no appchk-shell, but package-check needs the package unpacked
						$unpacked_pkgs{$path} = $temp_dir2;
					}
					process_dir($temp_dir2, $title_path);
				}
				else {
					if ($root) {
						push @failed_files, $title_path;
					}
					else {
						$journals{$path} = create_journal_INFO($path, $title_path, 'pkg');
						create_journal($path, 1, 'pkg', 'Failed to unpack the archive.');
					}
				}
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'pkg');
				create_journal($path, 101, 'pkg', 'No contents found in the DEB archive.');
			}
		}
	}
	elsif ($type eq 'elf-so') {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		push @list_appchk, $path;
		push @list_db, "$path|$title_path";
		print "PROGRESS: Add: ELF-so1.\n" if ($verbose);
		print "PROGRESS: Add: ELF-so2.\n" if ($verbose);
		print "PROGRESS: Add: APPDATA.\n" if ($verbose);
	}
	elsif ($type eq 'elf-bin') {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		push @list_appchk, $path;
		push @list_db, "$path|$title_path";
		print "PROGRESS: Add: ELF-bin.\n" if ($verbose);
		print "PROGRESS: Add: APPDATA.\n" if ($verbose);
	}
	elsif ($type eq 'elf-unsupported') {
		if ($root) {
			# The file was specified explicitly, we can't ignore it.
			# But it is unsupported by appchk, so, report failure.
			$journals{$path} = create_journal_INFO($path, $title_path, 'app');
			create_journal($path, 1, 'app', 'Unsupported ELF type.');
			print "PROGRESS: Add-processed: ELF.\n" if ($verbose);
		}
		# Else just skip the file.
	}
	elsif ($type eq 'perl') {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		push @list_appchk_pl, $path;
		push @list_db, "$path|$title_path";
		print "PROGRESS: Add: Perl.\n" if ($verbose);
		print "PROGRESS: Add: APPDATA.\n" if ($verbose);
	}
	elsif ($type eq 'python') {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		push @list_appchk_py, $path;
		push @list_db, "$path|$title_path";
		print "PROGRESS: Add: Python.\n" if ($verbose);
		print "PROGRESS: Add: APPDATA.\n" if ($verbose);
	}
	elsif ($type eq 'shell') {
		if ($appchecker_branch{'STD'} eq 'LSB') {
			$journals{$path} = create_journal_INFO($path, $title_path, 'app');
			push @list_appchk_sh, $path;
			push @list_db, "$path|$title_path";
			print "PROGRESS: Add: Shell.\n" if ($verbose);
			print "PROGRESS: Add: APPDATA.\n" if ($verbose);
		}
		else {
			# Shell scripts in Moblin AppChecker are not checked
			if ($root) {
				# The file was specified explicitly, we can't ignore it, so, report failure.
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 1, 'app', 'Checking shell scripts is not supported.');
			}
			# Else just skip the file.
		}
	}
	elsif ($type eq 'desktop') {
		if ($appchecker_branch{'STD'} ne 'LSB') {
			# Desktop files are checked in Moblin AppChecker only
			$journals{$path} = create_journal_INFO($path, $title_path, 'desktop');
			push @list_desktopchk, $path;
			print "PROGRESS: Add: DESKTOP.\n" if ($verbose);
		}
	}
	elsif ($type eq 'icon') {
		if ($appchecker_branch{'STD'} ne 'LSB') {
			# Desktop files and their icons are checked in Moblin AppChecker only
			push @list_desktopchk_icon, $path;
		}
	}
	elsif ($type =~ m/script:(.*)/) {
		my $script_type = $1;
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		create_journal($path, 1, 'app', 'Invalid script interpreter is used'.($script_type ? " ($script_type)" : '').'.');
		# Insert the unknown script into most-fit category
		# TODO: Maybe create additional category for such files?
		print "PROGRESS: Add-processed: ".(($appchecker_branch{'STD'} eq 'LSB') ? 'Shell' : 'Perl').".\n" if ($verbose);
	}
	elsif ($type =~ m/error:(.*)/) {
		if ($root) {
			push @failed_files, $title_path;
		}
		else {
			$journals{$path} = create_journal_INFO($path, $title_path, 'app');
			create_journal($path, 101, 'app', "Failed to read file contents.\n$1");
		}
	}
	elsif ($type eq 'unknown') {
		if ($root) {
			# The file was specified explicitly, we can't ignore it.
			# But we don't know how to test it, so, report failure.
			$journals{$path} = create_journal_INFO($path, $title_path, 'app');
			create_journal($path, 1, 'app', 'Failed to determine the file type.');
		}
		# Else just skip the file.
	}
	else {
		print "    Internal error: Bad return value from detect_file_type().\n    IN: '$path'\n    OUT: '$type'\n    Please report to lsb-appchk-support\@linuxfoundation.org.\n";
	}
}

# Create results directory
system(esc_cmdline("mkdir -p «$results_path/$results_dir/test_journals»"));
system(esc_cmdline("mkdir -p «$results_path/$results_dir/test_logs»"));
if ((!-d "$results_path/$results_dir/test_journals") or (!-d "$results_path/$results_dir/test_logs")) {
	$error = "Could not create subdirectories for storing the test results in\n$results_dir\n$!";
}
else {
	print "PROGRESS: Creating file list.\n" if ($verbose);
	# Create INFO files and fill file lists for testing
	foreach my $path (@paths) {
		process_file($path, $path, 1);
	}
	foreach my $pkg (@packages) {
		my @pkg_list;
		my $err_file = tmpnam();
		if (lc($package_manager) eq 'rpm') {
			system(esc_cmdline("rpm -q «$pkg» >«$err_file» 2>&1"));
			if (!$?) {
				@pkg_list = `rpm -ql $pkg`;
			}
			else {
				create_journal_INFO("PKG: $pkg", "PKG: $pkg|", 'pkg');
				create_journal("PKG: $pkg", 1, 'pkg', "Could not find package:\n".read_file($err_file));
			}
		}
		else {
			my $delim_pos = index($pkg, '_');
			my $pkg_name = (($delim_pos == -1) ? $pkg : substr($pkg, 0, $delim_pos));
			system(esc_cmdline("dpkg -l «$pkg_name» >«$err_file» 2>&1"));
			if (!$?) {
				@pkg_list = `dpkg -L $pkg_name`;
			}
			else {
				create_journal_INFO("PKG: $pkg", "PKG: $pkg|", 'pkg');
				create_journal("PKG: $pkg", 1, 'pkg', "Could not find package:\n".read_file($err_file));
			}
		}
		unlink($err_file);
		foreach (@pkg_list) {
			s/[\r\n]//g;
			process_file($_, "PKG: $pkg|$_", 0) if (!-d);
		}
	}
	# appchk-elf list
	my ($comp_list_fh, $comp_list_name) = tempfile("/tmp/$app_name.list.XXXXXX");
	if ($comp_list_fh) {
		foreach (@list_appchk) {
			print $comp_list_fh "$_\n";
		}
		close($comp_list_fh);
	}
	# appchk-shell list (Linux AppChecker only)
	my ($comp_list_sh_fh, $comp_list_sh_name) = tempfile("/tmp/$app_name.list_sh.XXXXXX");
	if ($comp_list_sh_fh) {
		foreach my $file (@list_appchk_sh) {
			my $journal = $journals{$file};
			print $comp_list_sh_fh "SH: $file\t$results_path/$results_dir/test_journals/$journal\n";
		}
		foreach my $file (@list_appchk) {
			print $comp_list_sh_fh "BIN: $file\n";
		}
		close($comp_list_sh_fh);
	}
	# appchk-perl list
	my ($comp_list_pl_fh, $comp_list_pl_name) = tempfile("/tmp/$app_name.list_pl.XXXXXX");
	if ($comp_list_pl_fh) {
		foreach my $file (@list_appchk_pl) {
			my $journal = $journals{$file};
			print $comp_list_pl_fh "$file\t$results_path/$results_dir/test_journals/$journal\n";
		}
		close($comp_list_pl_fh);
	}
	# compatibility analysis list
	my ($comp_list_db_fh, $comp_list_db_name) = tempfile("/tmp/$app_name.list_db.XXXXXX");
	if ($comp_list_db_fh) {
		foreach (@list_db) {
			print $comp_list_db_fh "$_\n";
		}
		close($comp_list_db_fh);
	}
	# pkgchk list (Moblin AppChecker only)
	my ($comp_list_pkg_fh, $comp_list_pkg_name) = tempfile("/tmp/$app_name.list_pkg.XXXXXX");
	if ($comp_list_pkg_fh) {
		if ($appchecker_branch{'STD'} ne 'LSB') {
			foreach my $file (@list_pkgchk) {
				print $comp_list_pkg_fh "$file\t$results_path/$results_dir/test_journals/$journals{$file}\t$unpacked_pkgs{$file}\n";
			}
		}
		close($comp_list_pkg_fh);
	}
	# desktopchk .desktop files list (Moblin AppChecker only)
	my ($comp_list_desktop_fh, $comp_list_desktop_name) = tempfile("/tmp/$app_name.list_desktop.XXXXXX");
	if ($comp_list_desktop_fh) {
		foreach my $file (@list_desktopchk) {
			print $comp_list_desktop_fh "$file\t$results_path/$results_dir/test_journals/$journals{$file}\n";
		}
		close($comp_list_desktop_fh);
	}
	# desktopchk icons list (Moblin AppChecker only)
	my ($comp_list_icon_fh, $comp_list_icon_name) = tempfile("/tmp/$app_name.list_icon.XXXXXX");
	if ($comp_list_icon_fh) {
		foreach (@list_desktopchk_icon) {
			print $comp_list_icon_fh "$_\n";
		}
		close($comp_list_icon_fh);
	}

	# Run tests
	if (scalar(@failed_files) == 0) {
		my $res;

		# pkgchk
		if ($appchecker_branch{'STD'} eq 'LSB') {
			# Linux AppChecker
			print "PROGRESS: Starting PKG.\n" if ($verbose);
			foreach my $file (@list_pkgchk) {
				my $filename = $file;
				$filename =~ s!^.*/([^/]+)$!$1!;
				print "PROGRESS: Process: PKG ($filename).\n" if ($verbose);
				$file =~ s/[\r\n]//g;
				system(esc_cmdline("«$tests_path/lsbpkgchk» -r $spec_version -j «$results_path/$results_dir/test_journals/$journals{$file}» -T$spec_profile «$file» >>«$tests_output» 2>&1"));
			}
		}
		else {
			# Moblin AppChecker
			$res = 0;
			print "PROGRESS: Starting PKG.\n" if ($verbose);
			if (open(PKGCHK, esc_cmdline("«$tests_path/moblinpkgchk.pl» -l «$comp_list_pkg_name» 2>&1 |"))) {
				my $line;
				my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
				while ($line = <PKGCHK>) {
					print TESTS_OUTPUT $line if ($out_opened);
					$line =~ s/[\r\n]//g;
					if ($verbose) {
						if ($line =~ m/^TESTING: (?:.*\/)?([^\/]+)$/) {
							print "PROGRESS: Process: PKG ($1).\n" if ($verbose);
						}
					}
				}
				close(TESTS_OUTPUT) if ($out_opened);
				close(PKGCHK);
				$res = $?;
			}
			else {
				system(esc_cmdline("echo \"Failed to start moblinpkgchk.pl: $!\" >>«$tests_output»"));
			}
			if ($res != 0) {
				if ($res & 127) {
					system(esc_cmdline("echo \"Error executing moblinpkgchk.pl (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>«$tests_output»"));
				}
				else {
					system(esc_cmdline("echo \"Unknown error: moblinpkgchk.pl exited with value ".($res >> 8)."!\" >>«$tests_output»"));
				}
			}
		}

		# appchk-shell
		if ($appchecker_branch{'STD'} eq 'LSB') {
			print "PROGRESS: Starting Shell.\n" if ($verbose);
			if (scalar(@list_appchk_sh) > 0) {
				$res = 0;
				if (open(APPCHK_SH, esc_cmdline("«$tests_path/lsbappchk-sh.pl» --lsb=$spec_version --list «$comp_list_sh_name» 2>&1 |"))) {
					my $line;
					my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
					while ($line = <APPCHK_SH>) {
						print TESTS_OUTPUT $line if ($out_opened);
						$line =~ s/[\r\n]//g;
						if ($verbose) {
							if ($line =~ m/^TESTING: (?:.*\/)?([^\/]+)$/) {
								print "PROGRESS: Process: Shell ($1).\n" if ($verbose);
							}
						}
					}
					close(TESTS_OUTPUT) if ($out_opened);
					close(APPCHK_SH);
					$res = $?;
				}
				else {
					system(esc_cmdline("echo \"Failed to start lsbappchk-sh: $!\" >>«$tests_output»"));
				}
				if ($res != 0) {
					if ($res & 127) {
						system(esc_cmdline("echo \"Error executing lsbappchk-sh (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>«$tests_output»"));
					}
					else {
						system(esc_cmdline("echo \"Unknown error: lsbappchk-sh exited with value ".($res >> 8)."!\" >>«$tests_output»"));
					}
				}
			}
		}

		# appchk-perl
		print "PROGRESS: Starting Perl.\n" if ($verbose);
		my $appchk_perl_tool = (($appchecker_branch{'STD'} eq 'LSB') ? 'lsbappchk.pl' : 'moblinappchk-perl.pl');
		if (scalar(@list_appchk_pl) > 0) {
			$res = 0;
			if (open(APPCHK_PL, esc_cmdline("«$tests_path/$appchk_perl_tool» -v $spec_version -j --list=«$comp_list_pl_name» --verbose 2>&1 |"))) {
				my $line;
				my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
				while ($line = <APPCHK_PL>) {
					print TESTS_OUTPUT $line if ($out_opened);
					$line =~ s/[\r\n]//g;
					if ($verbose) {
						if ($line =~ m/^TESTING: (?:.*\/)?([^\/]+)$/) {
							print "PROGRESS: Process: Perl ($1).\n" if ($verbose);
						}
					}
				}
				close(TESTS_OUTPUT) if ($out_opened);
				close(APPCHK_PL);
				$res = $?;
			}
			else {
				system(esc_cmdline("echo \"Failed to start appchk-perl: $!\" >>«$tests_output»"));
			}
			if ($res != 0) {
				if ($res & 127) {
					system(esc_cmdline("echo \"Error executing appchk-perl (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>«$tests_output»"));
				}
				else {
					system(esc_cmdline("echo \"Unknown error: appchk-perl exited with value ".($res >> 8)."!\" >>«$tests_output»"));
				}
			}
		}

		my $temp_dir = create_temp_dir('appcheck-journals');
		my $old_dir = cwd();
		chdir($temp_dir);

		# appchk-python
		my $appchk_python_tool = (($appchecker_branch{'STD'} eq 'LSB') ? 'lsbappchk.py' : 'moblinappchk-python.py');
		print "PROGRESS: Starting Python.\n" if ($verbose);
		foreach my $file (@list_appchk_py) {
			my $filename = $file;
			$filename =~ s!^.*/([^/]+)$!$1!;
			print "PROGRESS: Process: Python ($filename).\n" if ($verbose);
			$file =~ s/[\r\n]//g;
			system(esc_cmdline("«$tests_path/$appchk_python_tool» -v $spec_version -j «$file» >>«$tests_output» 2>&1"));
			my @journals_res = ();
			if (opendir(DIR, '.')) {
				@journals_res = grep(!m/^\.\.?$/, readdir(DIR));
				closedir(DIR);
				if (scalar(@journals_res) > 1) {
					print "    Warning: More than one journal were generated by appchk-python!\n    Only first one will be used.\n";
				}
				elsif (scalar(@journals_res) == 0) {
					print "    Warning: No journal was generated by appchk-python!\n";
				}
				else {
					system(esc_cmdline("mv «$journals_res[0]» «$results_path/$results_dir/test_journals/$journals{$file}» >>«$tests_output» 2>&1"));
				}
			}
			else {
				print "    Error: Failed to open directory $temp_dir for reading: $!";
			}
			system('rm -rf ./*');
		}
		chdir($old_dir);
		rmdir($temp_dir);

		# appchk-elf
		$res = 0;
		my $appchk_elf_tool = lc($appchecker_branch{'STD'}).'appchk';
		my $spec_profile_arg = (($appchecker_branch{'STD'} eq 'LSB') ? ' -T'.$spec_profile : '');
		if (open(APPCHK_ELF, esc_cmdline("«$tests_path/$appchk_elf_tool» -r $spec_version -j -o «$results_path/$results_dir/test_journals» -l «$comp_list_name»$spec_profile_arg 2>&1 |"))) {
			my $line;
			my %stages = ('SO-Pass1' => 'ELF-so1', 'SO-Pass2' => 'ELF-so2', 'BIN' => 'ELF-bin');
			my $stage = '';
			my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
			while ($line = <APPCHK_ELF>) {
				print TESTS_OUTPUT $line if ($out_opened);
				$line =~ s/[\r\n]//g;
				if ($verbose) {
					if ($line =~ m/^(SO-Pass1|SO-Pass2|BIN): .*\/([^\/]+)$/) {
						if ($stage ne $stages{$1}) {
							$stage = $stages{$1};
							print "PROGRESS: Starting $stage.\n" if ($verbose);
						}
						print "PROGRESS: Process: $stage ($2).\n" if ($verbose);
					}
				}
			}
			close(TESTS_OUTPUT) if ($out_opened);
			close(APPCHK_ELF);
			$res = $?;
		}
		else {
			system(esc_cmdline("echo \"Failed to start appchk: $!\" >>«$tests_output»"));
		}
		if ($res != 0) {
			if ($res & 127) {
				system(esc_cmdline("echo \"Error executing appchk (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>«$tests_output»"));
			}
			else {
				system(esc_cmdline("echo \"Unknown error: appchk exited with value ".($res >> 8)."!\" >>«$tests_output»"));
			}
		}

		# moblindesktopchk
		if ($appchecker_branch{'STD'} ne 'LSB') {
			print "PROGRESS: Starting DESKTOP.\n" if ($verbose);
			if (scalar(@list_desktopchk) > 0) {
				$res = 0;
				if (open(DESKTOPCHK, esc_cmdline("«$tests_path/moblindesktopchk.py» -j «$results_path/$results_dir/test_journals» -d «$comp_list_desktop_name» -i «$comp_list_icon_name» 2>&1 |"))) {
					my $line;
					my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
					while ($line = <DESKTOPCHK>) {
						print TESTS_OUTPUT $line if ($out_opened);
						$line =~ s/[\r\n]//g;
						if ($verbose) {
							if ($line =~ m/^TESTING: (?:.*\/)?([^\/]+)$/) {
								print "PROGRESS: Process: DESKTOP ($1).\n" if ($verbose);
							}
						}
					}
					close(TESTS_OUTPUT) if ($out_opened);
					close(DESKTOPCHK);
					$res = $?;
				}
				else {
					system(esc_cmdline("echo \"Failed to start moblindesktopchk: $!\" >>«$tests_output»"));
				}
				if ($res != 0) {
					if ($res & 127) {
						system(esc_cmdline("echo \"Error executing moblindesktopchk (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>«$tests_output»"));
					}
					else {
						system(esc_cmdline("echo \"Unknown error: moblindesktopchk exited with value ".($res >> 8)."!\" >>«$tests_output»"));
					}
				}
			}
		}

		print "PROGRESS: Starting JOURNAL.\n" if ($verbose);

		# Save information about test run
		system(esc_cmdline("touch «$results_path/HISTORY»"));
		if (open(FILE, "+<$results_path/HISTORY")) {
			flock(FILE, LOCK_EX);
			seek(FILE, 0, SEEK_END);
			print FILE "$results_dir\n";
			flock(FILE, LOCK_UN);
			close(FILE);
		}
		else {
			system(esc_cmdline("echo \"Could not open HISTORY file!\nThe test results will not be listed on the Results page.\" >>«$tests_output» 2>&1"));
		}
		if (open(COMMENTS_FILE, ">$results_path/$results_dir/test_logs/test_comments")) {
			print COMMENTS_FILE $comments;
			close(COMMENTS_FILE);
		}
		system(esc_cmdline("echo \"$spec_version\" >> «$results_path/$results_dir/test_logs/spec_version»"));
		system(esc_cmdline("echo \"$architecture\" >> «$results_path/$results_dir/test_logs/host_platform»"));
		system(esc_cmdline("cp «$utils_path/ManualTests» «$results_path/$results_dir/» >>«$tests_output» 2>&1"));

		# Generate the report
		if (-s "$results_path/$results_dir/test_logs/tests_output" != 0) {
			system(esc_cmdline("echo \"\" >>«$tests_output»"));
		}
		$res = 0;
		if (open(GEN_REPORT, esc_cmdline("«$utils_path/report.pl» «$results_path/$results_dir» -n «$app_name» -o «$results_dir.tgz» 2>&1 |"))) {
			my $line;
			my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
			while ($line = <GEN_REPORT>) {
				print TESTS_OUTPUT $line if ($out_opened);
				$line =~ s/[\r\n]//g;
				if ($line =~ m/^Reading journal: test_journals\/(.*)$/) {
					print "PROGRESS: Process: JOURNAL ($1).\n" if ($verbose);
				}
			}
			close(TESTS_OUTPUT) if ($out_opened);
			close(GEN_REPORT);
			$res = $?;
		}
		else {
			system(esc_cmdline("echo \"Failed to start report.pl: $!\" >>«$tests_output»"));
		}
		if ($res != 0) {
			if ($res & 127) {
				system(esc_cmdline("echo \"Error executing report.pl (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>«$tests_output»"));
			}
			else {
				system(esc_cmdline("echo \"Unknown error: report.pl exited with value ".($res >> 8)."!\" >>«$tests_output»"));
			}
		}

		# Collects application data
		# apptodb needs PM file located in its directory.
		print "PROGRESS: Starting APPDATA.\n" if ($verbose);
		chdir($utils_path);
		system(esc_cmdline("echo \"\" >>«$tests_output»"));
		system(esc_cmdline("echo \"Collecting data about application files\" >>«$tests_output»"));
		$res = 0;
		if (open(APPTODB, esc_cmdline("«$utils_path/apptodb.pl» --verbose --trsp-symlink -a «$architecture» -l «$comp_list_db_name» -o «$upload_data_file» «$app_name» 2>&1 |"))) {
			my $line;
			my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
			while ($line = <APPTODB>) {
				print TESTS_OUTPUT $line if ($out_opened);
				$line =~ s/[\r\n]//g;
				if ($line =~ m/^Processing file: '.*\/([^\/]+)'$/) {
					print "PROGRESS: Process: APPDATA ($1).\n" if ($verbose);
				}
			}
			close(TESTS_OUTPUT) if ($out_opened);
			close(APPTODB);
			$res = $?;
		}
		else {
			system(esc_cmdline("echo \"Failed to start apptodb.pl: $!\" >>«$tests_output»"));
		}
		if ($res != 0) {
			if ($res & 127) {
				system(esc_cmdline("echo \"Error executing apptodb.pl (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>«$tests_output»"));
			}
			else {
				system(esc_cmdline("echo \"Unknown error: apptodb.pl exited with value ".($res >> 8)."!\" >>«$tests_output»"));
			}
		}

		# Generates application analysis files
		print "PROGRESS: Starting MERGE.\n" if ($verbose);

		# Analysis
		system(esc_cmdline("«$utils_path/analysis.pl» «$spec_version» «$architecture» «$dbdata_path» «$upload_data_file» >>«$tests_output» 2>&1"));

		# Create TAR.GZ
		my $full_result_archive = "$results_path/$results_dir/$results_dir.tgz";
		if (-e $full_result_archive) {
			unlink($full_result_archive);
			if (-f $full_result_archive) {
				die "Failed to remove old archive.\n";
			}
		}
		system(esc_cmdline("(cd «$results_path» && tar czf «$results_dir.tgz» «$results_dir/» --force-local && mv «$results_dir.tgz» «$results_dir/») >>«$tests_output» 2>&1"));

		print "PROGRESS: Starting CLEANUP.\n" if ($verbose);
	}
	else {
		$error = "Could not process the following files:\n".join("\n", @failed_files);
	}
	unlink($comp_list_name);
	unlink($comp_list_sh_name);
	unlink($comp_list_pl_name);
	unlink($comp_list_pkg_name);
	unlink($comp_list_desktop_name);
	unlink($comp_list_icon_name);
	unlink($comp_list_db_name);
	foreach (@clean_dirs) {
		system(esc_cmdline("rm -rf «$_» >/dev/null 2>&1"));
	}
	print "PROGRESS: Starting FINISH.\n" if ($verbose);
}

if (defined($error)) {
	system(esc_cmdline("rm -rf «$results_path/$results_dir» >/dev/null 2>&1"));
	print "ERROR: ".join("\nERROR: ", split(/\n/, $error))."\n";
	exit 4;
}

print "RESULT: $results_dir\n";
exit 0;
